/*
  IfOnline.c - Only execute an http request if online.

  Jason Hood, 9, 23 & 28 February, 6 to 10 March, 2002.

  Prevent the browser from opening if you are offline.
*/


#include <windows.h>
#include <shellapi.h>
#include <ddeml.h>
#include <stdlib.h>
#include <string.h>
#include "IfOnline.h"

#define HTTP "http\\shell\\open\\command"

#ifndef IDC_LINK
#define IDC_LINK MAKEINTRESOURCE( 32649 )  // Why isn't this already defined?
#endif


LPTSTR argument( LPTSTR cmdtail, LPTSTR arg, LPCTSTR parm );

BOOL CALLBACK OfflineDlg( HWND, UINT, WPARAM, LPARAM );
BOOL CALLBACK AboutDlg( HWND, UINT, WPARAM, LPARAM );


// Return TRUE if online, FALSE if off.
BOOL online( void )
{
  HKEY	remote;
  DWORD on = FALSE, len;

  if (RegOpenKeyEx( HKEY_LOCAL_MACHINE,
		    "System\\CurrentControlSet\\Services\\RemoteAccess", 0,
		    KEY_READ, &remote ) == ERROR_SUCCESS)
  {
    len = sizeof(DWORD);
    RegQueryValueEx( remote, "Remote Connection", NULL, NULL, &on, &len );
    RegCloseKey( remote );
  }

  return on;
}


// Return TRUE if successfully installed, FALSE if not (no default browser).
BOOL install( void )
{
  HKEY	http;
  TCHAR command[1024];
  DWORD len;

  if (RegOpenKeyEx( HKEY_CLASSES_ROOT, HTTP, 0,
		    KEY_ALL_ACCESS, &http ) != ERROR_SUCCESS)
  {
    MessageBox( NULL, "There is no http association.", PNAME, MB_OK );
    return FALSE;
  }

  len = sizeof(command);
  RegQueryValueEx( http, NULL, NULL, NULL, command, &len );

  // Don't set the browser to IfOnline, if it's already installed.
  // Let's assume no-one will rename it, or a browser will be called it.
  if (stristr( command, PNAME ) == NULL)
    RegSetValueEx( http, "Online", 0, REG_SZ, command, len );

  // However, always set this, in case it's moved.
  command[0] = '"';
  len = GetModuleFileName( NULL, command+1, sizeof(command)-5 );
  lstrcpy( command + len + 1, "\" %1" );
  RegSetValueEx( http, NULL, 0, REG_SZ, command, len + 6*sizeof(TCHAR) );

  RegCloseKey( http );

  return TRUE;
}


// Return TRUE if successfully uninstalled, FALSE if not installed.
BOOL uninstall( void )
{
  HKEY	http;
  TCHAR command[1024];
  DWORD len;

  len = sizeof(command);
  if (RegOpenKeyEx( HKEY_CLASSES_ROOT, HTTP, 0,
		    KEY_ALL_ACCESS, &http ) != ERROR_SUCCESS
      || RegQueryValueEx( http, "Online", NULL, NULL,
			  command, &len ) != ERROR_SUCCESS)
  {
    MessageBox( NULL, "Not installed.", PNAME, MB_OK );
    return FALSE;
  }

  RegSetValueEx( http, NULL, 0, REG_SZ, command, len );
  RegDeleteValue( http, "Online" );

  RegCloseKey( http );

  return TRUE;
}


DWORD idInst = 0;
HSZ   hszService, hszTopic;
PTSTR ddecommand = NULL;
int   ddecmd_len;

BOOL ddeopen( void );
void ddeclose( void );


// Intercept the browser's open url transaction.
HDDEDATA CALLBACK DDECallback( UINT uType, UINT uFmt, HCONV hconv,
			       HSZ hsz1, HSZ hsz2, HDDEDATA hdata,
			       DWORD dwData1, DWORD dwData2 )
{
  switch (uType)
  {
    case XTYP_CONNECT:
      return (HDDEDATA)(hsz1 == hszTopic);	// Only the one service
    break;

    case XTYP_EXECUTE:
      ddecmd_len = DdeGetData( hdata, NULL, 0, 0 );
      ddecommand = malloc( ddecmd_len );
      DdeGetData( hdata, ddecommand, ddecmd_len, 0 );
      return (HDDEDATA)DDE_FACK;
    break;
  }
  return NULL;
}


// Initialise the DDEML and fake the browser service.
// Return TRUE if successful, FALSE if the browser doesn't use DDE.
BOOL ddeopen( void )
{
  HKEY	ddekey, ddeval;
  DWORD len;
  TCHAR ddestr[1024];

  if (RegOpenKeyEx( HKEY_CLASSES_ROOT, "http\\shell\\open\\ddeexec", 0,
		    KEY_READ, &ddekey ) != ERROR_SUCCESS)
  {
    return FALSE;
  }

  if (DdeInitialize( &idInst, DDECallback,
		     CBF_FAIL_ADVISES | CBF_FAIL_POKES | CBF_FAIL_REQUESTS
		     | CBF_SKIP_ALLNOTIFICATIONS, 0L ))
  {
    MessageBox( NULL, "Unable to initialise DDE.", PNAME, MB_OK );
    RegCloseKey( ddekey );
    return FALSE;
  }

  RegOpenKeyEx( ddekey, "Application", 0, KEY_READ, &ddeval );
  len = sizeof(ddestr);
  RegQueryValueEx( ddeval, NULL, NULL, NULL, ddestr, &len );
  RegCloseKey( ddeval );
  hszService = DdeCreateStringHandle( idInst, ddestr, CP_WINANSI );

  RegOpenKeyEx( ddekey, "Topic", 0, KEY_READ, &ddeval );
  len = sizeof(ddestr);
  RegQueryValueEx( ddeval, NULL, NULL, NULL, ddestr, &len );
  RegCloseKey( ddeval );
  hszTopic = DdeCreateStringHandle( idInst, ddestr, CP_WINANSI );

  RegCloseKey( ddekey );

  if (DdeNameService( idInst, hszService, 0, DNS_REGISTER ) == 0)
  {
    MessageBox( NULL, "Unable to register DDE service.", PNAME, MB_OK );
    ddeclose();
    return FALSE;
  }

  return TRUE;
}


// Pass along the DDE transaction to the actual browser.
void ddeexec( void )
{
  HCONV hConv;
  int	tries;

  tries = 0;
  while ((hConv = DdeConnect( idInst, hszService, hszTopic, NULL )) == NULL
	 && tries++ < 20)
  {
    Sleep( 500 );
  }
  if (hConv == NULL)
  {
    MessageBox( NULL, "Unable to create DDE conversation.", PNAME, MB_OK );
  }
  else
  {
    DdeClientTransaction( ddecommand, ddecmd_len, hConv, 0, 0, XTYP_EXECUTE,
			  500, NULL );
    DdeDisconnect( hConv );
  }

  ddeclose();
}


void ddeclose( void )
{
  DdeFreeStringHandle( idInst, hszService );
  DdeFreeStringHandle( idInst, hszTopic );
  DdeUninitialize( idInst );

  if (ddecommand != NULL)
    free( ddecommand );
}


int WINAPI WinMain( HINSTANCE hInst, HINSTANCE hPrev,
		    LPSTR lpCmdLine, int nCmdShow )
{
  HKEY	 http;
  TCHAR  command[1024], arg[1024];
  DWORD  len;
  PTSTR  exe;
  DWORD  on;
  BOOL	 dde;

  if (*lpCmdLine == 0)
  {
    DialogBox( NULL, MAKEINTRESOURCE( DLG_ABOUT ), NULL, AboutDlg );
    return 0;
  }

  if (lpCmdLine[0] == '-')
  {
    if (lpCmdLine[1] == 'i')
    {
      install();
    }
    else if (lpCmdLine[1] == 'u')
    {
      uninstall();
    }
    return 0;
  }

  dde = ddeopen();

  if (!online())
  {
    on = DialogBoxParam( NULL, MAKEINTRESOURCE( DLG_OFFLINE ), NULL,
			 OfflineDlg, (LPARAM)lpCmdLine );
    if (on == ID_COPY)
    {
      HGLOBAL url = GlobalAlloc( GMEM_DDESHARE, strlen( lpCmdLine ) + 1 );
      char* url_str = GlobalLock( url );
      strcpy( url_str, lpCmdLine );
      GlobalUnlock( url );
      if (OpenClipboard( NULL ))
      {
	EmptyClipboard();
	SetClipboardData( CF_TEXT, url );
	CloseClipboard();
      }
      else
	MessageBox( NULL, "Unable to open the clipboard.", PNAME, MB_OK );
    }
    if (on != IDOK)
    {
      if (dde) ddeclose();
      return 0;
    }
  }

  if (RegOpenKeyEx( HKEY_CLASSES_ROOT, HTTP, 0,
		    KEY_READ, &http ) != ERROR_SUCCESS)
  {
    MessageBox( NULL, "There is no http association.", PNAME, MB_OK );
    return 0;
  }
  len = sizeof(command);
  RegQueryValueEx( http, "Online", NULL, NULL, command, &len );
  RegCloseKey( http );

  exe = argument( command, arg, lpCmdLine );

  if (dde) DdeNameService( idInst, hszService, 0, DNS_UNREGISTER );
  ShellExecute( NULL, NULL, exe, arg, NULL, SW_SHOWNORMAL );
  if (dde) ddeexec();

  return 0;
}


// Separate the program from the argument, substituting parm for %1 in the arg.
LPTSTR argument( LPTSTR cmdtail, LPTSTR arg, LPCTSTR parm )
{
  LPTSTR exe;
  TCHAR  term;

  term = (*cmdtail == '\"') ? (++cmdtail, '\"') : ' ';
  exe = cmdtail;
  while (*cmdtail && *cmdtail != term)
    ++cmdtail;

  if (*cmdtail)
  {
    *cmdtail++ = 0;
    if (*cmdtail == ' ') ++cmdtail;

    while (*cmdtail)
    {
      if (*cmdtail == '%')
      {
	++cmdtail;
	if (*cmdtail == '1')
	{
	  arg += wsprintf( arg, "%s", parm );
	}
	else if (*cmdtail == '%')
	{
	  *arg++ = '%';
	}
	++cmdtail;
      }
      else *arg++ = *cmdtail++;
    }
  }
  *arg = 0;

  return exe;
}


BOOL CALLBACK OfflineDlg( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
  switch (msg)
  {
    case WM_INITDIALOG:
    {
      HDC  dc;
      SIZE size;
      HWND url;
      RECT rect;
      int  wid;

      // Resize the dialog to the width of the url.
      // Note: if I use the commented font in the resource, the extent still
      //       appears to use the system font.
      url = GetDlgItem( hwnd, ID_URL );
      dc = GetDC( url );
      GetTextExtentPoint32( dc, (LPCTSTR)lParam, lstrlen( (LPCTSTR)lParam ),
			    &size );
      ReleaseDC( url, dc );
      size.cy = size.cx + 2 * (GetSystemMetrics( SM_CXEDGE )
			       + 6 * LOWORD( GetDialogBaseUnits() ) / 4);

      GetWindowRect( hwnd, &rect );
      wid = rect.right - rect.left + 1;
      if (size.cy > wid)
      {
	int adjust = (size.cy - wid) / 2;
	POINT p;
	int id[] = { IDOK, ID_COPY, IDCANCEL };
	int j;

	MoveWindow( hwnd, rect.left - adjust, rect.top,
			  size.cy, rect.bottom - rect.top + 1, FALSE);
	GetWindowRect( url, &rect );
	p.x = rect.left;
	p.y = rect.top;
	ScreenToClient( hwnd, &p );
	MoveWindow( url, p.x, p.y, size.cx, rect.bottom - rect.top + 1, FALSE );

	// Reposition the three buttons, so they remain centered.
	for (j = 0; j < 3; ++j)
	{
	  HWND w = GetDlgItem( hwnd, id[j] );
	  GetWindowRect( w, &rect );
	  p.x = rect.left + adjust;
	  p.y = rect.top;
	  ScreenToClient( hwnd, &p );
	  SetWindowPos( w, NULL, p.x, p.y, 0, 0,
			SWP_NOACTIVATE | SWP_NOOWNERZORDER | SWP_NOREDRAW |
			SWP_NOSIZE | SWP_NOZORDER );
	}
      }

      SendMessage( url, WM_SETTEXT, 0, lParam );
    }
    break;

    case WM_COMMAND:
      switch (LOWORD( wParam ))
      {
	case IDOK:
	case ID_COPY:
	case IDCANCEL:
	  EndDialog( hwnd, LOWORD( wParam ));
	break;

	default: return FALSE;
      }
    break;

    case WM_CLOSE:
      EndDialog( hwnd, IDCANCEL );
    break;

    default: return FALSE;
  }
  return TRUE;
}


// Display author, version and copyright info.
// It can also send me an e-mail or go to my homepage.
BOOL CALLBACK AboutDlg( HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
  HCURSOR hand;
  HWND	  link;
  LONG	  id;

  switch (msg)
  {
    case WM_INITDIALOG:
      hand = LoadCursor( NULL, IDC_LINK	);
      link = GetDlgItem( hwnd, ID_AUTHOR );
      SetClassLong( link, GCL_HCURSOR, (LONG)hand );
      //link = GetDlgItem( hwnd, ID_HOME );
      //SetClassLong( link, GCL_HCURSOR, (LONG)hand );
    break;

    case WM_CTLCOLORSTATIC:
      id = GetWindowLong( (HWND)lParam, GWL_ID );
      if (id == ID_AUTHOR || id == ID_HOME)
      {
	SetTextColor( (HDC)wParam, RGB( 0, 0, 255 ) );
	SetBkMode( (HDC)wParam, TRANSPARENT );
	return (BOOL)GetSysColorBrush( COLOR_3DFACE );
      }
      return 0;
    break;

    case WM_COMMAND:
      switch (LOWORD( wParam ))
      {
	case IDOK: EndDialog( hwnd, 0 ); break;

	case ID_AUTHOR:
	case ID_HOME:
	  if (HIWORD( wParam ) == STN_CLICKED)
	    ShellExecute( hwnd, NULL, (LOWORD( wParam ) == ID_AUTHOR) ?
			  "mailto:jadoxa@yahoo.com.au?Subject="PNAME" "PVERS :
			  "http://adoxa.tsx.org/", NULL, ".", 0 );
	break;

	default: return 0;
      }
      break;

    case WM_CLOSE:
      EndDialog( hwnd, 0 );
    break;

    default: return 0;
  }
  return 1;
}
